// BTree.h: interface for the BTree class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_BTREE_H__FF772CA7_A227_4121_BB56_6ACBF7B75002__INCLUDED_)
#define AFX_BTREE_H__FF772CA7_A227_4121_BB56_6ACBF7B75002__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#ifndef __BTREE__H_
#define __BTREE__H_

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ INCLUDES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#include <vector>
#include <fstream>
#include <string>
#include "Buffer.h"
#include "AvailList.h"
#include <list>
class BTNCell;
class BTNode;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ TYPES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
typedef std::string		STRING;
typedef std::list<int>	INTLIST;
typedef std::streamoff	OFFSET;
#define NODESIZE		4 
#define NODEHALF		(NODESIZE / 2)

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CLASS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class BTree
{
public:
	BTree( Cache * cache, char type);
	~BTree();

	void			add( STRING key, int ObjectID);
	void			remove( STRING key);

	STRING			getplacestring( OFFSET location);
	void			deleteplacestring( OFFSET location, int size);
	OFFSET			putplacestring( STRING newdata);

	char			getfieldterm(){return buf.getfieldterm();}
	char			getrecordterm(){return buf.getrecordterm();}

	void			find( STRING key);

	void			out();

	INTLIST			lastfounds;

private:
	BTNCell			*root;
	BTreeBuffer		buf;
	AvailHandler	avail;
};

class BTNCell
{
	friend class BTree;
	friend class BTNode;
public:
	BTNCell( BTree * mypapa , STRING key = "", BTNode *point = NULL, OFFSET byteoffset = -1);
	BTNCell( BTree * mypapa , STRING key, int ObjectID);
	~BTNCell();

	void		BTNCell::swap( BTNCell *& first, BTNCell *& second);
	void		merge( BTNCell * first, BTNCell * second);
	void		split( BTNCell * first, BTNCell * second);
	void		cure();

	STRING		getKey(){return key;}
	BTNode*		getReference(){return reference;}
	int			getObjectID(){return ObjectID;}
	OFFSET		getByteOffset(){return byteoffset;}

private:
	char		operator <(const BTNCell & another){return (key < another.key);}
	char		operator!=(const BTNCell & another){return (key != another.key);}

	STRING		key;
	BTNode		*reference;
	BTree		*mypapa;
	char		recursive;
	
	union
	{
		OFFSET	byteoffset;
		int		ObjectID;
	};
};

class BTNode
{
	friend class BTree;
	friend class BTNCell;
public:
	BTNode( BTNCell * first, BTNCell * second, BTree * mypapa, char flag = 0);
	BTNode( BTree * mypapa, char flag = 0);
	~BTNode();

	void					find( STRING key);

	int						merge( BTNode * first, BTNode * second);
	void					split( BTNode * first, BTNode * second);

	STRING					bigest();
	char					add( STRING key, int ObjectID);
	char					remove( STRING key);
	STRING					pack( char fterm, char rterm);
	void					unpack( BTreeBuffer & buf, int offset, char fterm, char rterm);

	void			out();

private:	

	int						label;
	char					flag;		//leaf -> 1, branch -> 0
	std::vector<BTNCell *>	mynodes;
	char					recursive;
	BTree					*mypapa;
};


#endif
#endif // !defined(AFX_BTREE_H__FF772CA7_A227_4121_BB56_6ACBF7B75002__INCLUDED_)
